/*
Critical Mass
Board class implementation
Copyright(C) 2002, Jem Berkes <umberkes at cc dot umanitoba dot ca>
*/


#include "defs.h"
#include "board.h"
#include <memory.h>


//////////////////////////////////////
//    PUBLIC FUNCTIONS (EXTERNAL)   //
//////////////////////////////////////


/*
 Destructor
*/
the_board::~the_board()
{ destroy(); }


/*
 Default constructor
*/
the_board::the_board()
{
width = 0;
height = 0;
contents = NULL;
}


/*
 Constructor
*/
the_board::the_board(short thewidth, short theheight)
{
if (thewidth < 1)
	width = 1;
else
	width = thewidth;
if (theheight < 1)
	height = 1;
else
	height = theheight;
build();
}


/*
 Copy constructor
*/
the_board::the_board(const the_board& other)
{
contents = NULL;
copy_from(other);
}


/*
 Overloaded assignment operator
*/
the_board& the_board::operator = (const the_board& other)
{
if (this != &other)
	copy_from(other);
return *this;
}


/*
 Read contents of a cell
*/
uchar the_board::read_cell(short x, short y)
{ return *get_cell(x, y); }


/*
 Move a player into a cell. Returns whether move succeeded.
*/
bool the_board::move_into_cell(uchar player, short x, short y)
{
uchar* cell = get_cell(x, y);
if (cell == NULL)
	return false;
uchar owner = *cell & PLAYER_MASK;	// bitwise: extract owner
if ( (owner != 0) && (owner != player) )
	return false;	// move fails
int total_cells = width*height;
// whoever was the last move, no longer is
for (int i=0; i<total_cells; i++)
	contents[i] &= ~LASTMOVE_MASK;	// clear last move flags
uchar bombs = *cell & BOMB_MASK;	// bitwise: extract bombs
bombs++;
// construct cell: which player, how many bombs, and last move flag
*cell = player | bombs | LASTMOVE_MASK;
if (bombs == cell_max_bombs(x, y))
	explode_cells();
return true;
}


/*
 How many bombs does it take to make this cell critical?
*/
uchar the_board::cell_max_bombs(short x, short y)
{
uchar max = 4;	// start with maximum of 4, then adjust
if (x == 0)
	max--;
if (y == 0)
	max--;
if (x == width-1)
	max--;
if (y == height-1)
	max--;
return max;
}


/*
 Has anybody won the board yet?
*/
uchar the_board::who_won()
{
uchar aplayer = 0;
uchar* cell;

for (int i=0; i<width; i++)
	{
	for (int j=0; j<height; j++)
		{
		cell = get_cell(i, j);
		uchar owner = *cell & PLAYER_MASK;	// bitwise: extract owner
		if (owner != 0)
			{
			if (aplayer == 0)
				aplayer = owner;
			else if (aplayer != owner)
				return 0;
			}
		}
	}
return aplayer;
}


/*
 Find out what the last move was
*/
bool the_board::read_lastmove(int& x, int& y)
{
for (int i=0; i<width; i++)
	{
	for (int j=0; j<height; j++)
		{
		uchar cell = read_cell(i, j);
		if ( (cell&LASTMOVE_MASK) == LASTMOVE_MASK)
			{
			x = i;
			y = j;
			return true;
			}
		}
	}
return false;
}


/*
 Player makes random move
*/
void the_board::random_move(uchar player)
{
int x, y;
do	{
	x = rand() % width;
	y = rand() % height;
	} while (!move_into_cell(player, x, y));
}


/*
 Make an intelligent computer move
*/
void the_board::smart_move(uchar player)
{
int depth = 0;
uchar opponent = PLAYER2;
if (player == PLAYER2)
	opponent = PLAYER1;
the_board best_board = find_best_move(player, opponent, depth);
copy_from(best_board); // copy best result to this
}


//////////////////////////////////////
//   PRIVATE FUNCTIONS (INTERNAL)   //
//////////////////////////////////////


/*
 Copy contents from another board to this
*/
void the_board::copy_from(const the_board& other)
{
destroy();	// cleanup self
if (other.contents != NULL)
	{
	width = other.width;
	height = other.height;
	contents = new uchar[width*height];
	// copy a memory block (cells are a simple array)
	memcpy(contents, other.contents, (width*height));
	}
}


/*
 Build 2 dimensional board into 1 dimensional array
*/
void the_board::build()
{
contents = new uchar[width*height];
memset(contents, 0, (width*height));	// Initialize to 0 (blank cell)
}


/*
 Destroy the board and de-allocate its memory
*/
void the_board::destroy()
{
if (contents != NULL)
	delete [] contents;
contents = NULL;
width = 0;
height = 0;
}


/*
 Returns pointer to a specific cell
*/
uchar* the_board::get_cell(short x, short y)
{
if ( (x>=0) && (y>=0) && (x<width) && (y<height) )
	return &contents[(y*width) + x];
else
	return NULL;
}


/*
 Occupy a cell during critical mass explosions
*/
void the_board::occupy_cell(short x, short y, uchar player)
{
if ( (x>=0 && x<width) && (y>=0 && y<height) )
	{
	uchar* cell = get_cell(x, y);
	// Encode new owner, increase number of bombs
	uchar bombs = *cell & BOMB_MASK;
	bombs++;
	*cell &= LASTMOVE_MASK;	// preserve last move flag
	*cell |= player | bombs;
	}
}


/*
 Explode cells (automatic), iterative
*/
void the_board::explode_cells()
{
int i, j;
uchar* cell;
bool critical;

do	{
	critical = false;
	// Update criticality of each cell
	for (i=0; i<width; i++)
		{
		for (j=0; j<height; j++)
			{
			cell = get_cell(i, j);
			if ( (*cell & BOMB_MASK) >= cell_max_bombs(i, j))
				{
				*cell |= GONE_CRITICAL;
				critical = true;
				}
			}
		}

	if (critical)
		{
		// Explode all critical cells
		for (i=0; i<width; i++)
			{
			for (j=0; j<height; j++)
				{
				cell = get_cell(i, j);
				if ( (*cell & GONE_CRITICAL) == GONE_CRITICAL)
					{
					uchar owner = *cell & PLAYER_MASK;
					*cell &= LASTMOVE_MASK;	// Preserve last move flag
					occupy_cell(i-1, j, owner);
					occupy_cell(i+1, j, owner);
					occupy_cell(i, j-1, owner);
					occupy_cell(i, j+1, owner);					
					}
				}
			}
		}
	} while (critical && (who_won()==0) );
}


/*
 Board evaluator. More positive, better for player 1.
 More negative, better for player 2.

 Note: Having more bombs in a cell is not necessarily an advantage
 (it could also work against), so this is not weighed into calculation
*/
short the_board::evaluate()
{
short score = 0;
uchar* cell;
for (int i=0; i<width; i++)
	{
	for (int j=0; j<height; j++)
		{
		cell = get_cell(i, j);
		uchar owner = *cell & PLAYER_MASK;
		int points = 1;	// default points for a square
		// Now apply edge bonus(es). Most points for corner.
		if (i==0)
			points += 2;
		else if (i==width-1)
			points += 2;
		if (j==0)
			points += 2;
		else if (j==height-1)
			points += 2;
		// Note that being on two edges (i.e. a corner) is a HUGE bonus

		if (owner == PLAYER1)
			score += points;
		else if (owner == PLAYER2)
			score -= points;
		}
	}
return score;
}


/*
 Recursive use of game trees. Find the best move that this player can
 make. Rate limit keeps game tree from growing too large. Returns the
 resulting board from this "best move"
*/
the_board the_board::find_best_move(uchar myself, uchar opponent, int &depth)
{
the_board best_move = *this; // We don't have a move yet; current state
bool changed = false;

if (depth < MAX_TREE_DEPTH)
	{
	depth++;	// Gone deeper into game tree
	short best_result = evaluate(); // Best result so far (no change)

	for (int x=0; x<width; x++)
		{
		for (int y=0; y<height; y++)
			{
			// FOR EVERY POSSIBLE MOVE

			// Initialise the possible move
			the_board possible = *this;
			// Try making the move into (x, y)
			if (possible.move_into_cell(myself, x, y))
				{
				// If the proposed move worked, check for quick win
				if (possible.who_won() == myself)
					{
					depth--;
					return possible; // we've won, so this is best move
					}
				else
					{
					// Find strongest possible response to this move
					the_board result =
						possible.find_best_move(opponent, myself, depth);
					short value = result.evaluate(); // evaluate the result
					// 	If the result is in my favour more than best_result
					if  ( ((myself==PLAYER1) && (value>best_result)) ||
						  ((myself==PLAYER2) && (value<best_result)) )
						{
						best_result = value; // store best result
						best_move = possible; // store best move
						changed = true;
						}
					}
				}
			}
		}
	depth--;
	}

if (!changed)	// if I couldn't find a good move to make
	best_move.random_move(myself); // make a random move anyway
return best_move;
}
